//     This file is part of cdloop
//     Copyright (C) 2001-2003 Claus Brunzema <mail@cbrunzema.de>
// 
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
// 
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program; if not, write to the Free Software
//     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// 
#include "CDDrive.hh"

#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <mntent.h>
#include <sys/types.h>
typedef unsigned long long __u64;
#include <linux/cdrom.h>

static CDAbsPos tocentry_to_cdpos(const struct cdrom_tocentry& e)
{
    CDAbsPos pos;
    pos.minute=e.cdte_addr.msf.minute;
    pos.second=e.cdte_addr.msf.second;
    pos.frame=e.cdte_addr.msf.frame;
    return pos;
}

void
CDDrive::
update_disc()
{
    disc.clear();

    struct cdrom_tochdr h;
    struct cdrom_tocentry e;
    if(ioctl(cd_fd, CDROMREADTOCHDR, &h)) return;

    int n_track=h.cdth_trk1;
    if(n_track<1) return;

    int track=1;
    bool is_data=false;
    do {
        e.cdte_track=track;
        e.cdte_format=CDROM_MSF;
        if(ioctl(cd_fd, CDROMREADTOCENTRY, &e)) return;
        is_data=(e.cdte_ctrl & CDROM_DATA_TRACK);
        disc.push_back(tocentry_to_cdpos(e));
        track++;
    }while(track<=n_track && !is_data);
    if(!is_data) {
        e.cdte_track=CDROM_LEADOUT;
        e.cdte_format=CDROM_MSF;
        if(ioctl(cd_fd,CDROMREADTOCENTRY,&e)) return;
        disc.push_back(tocentry_to_cdpos(e));
    }
    stop();
}

CDDrive::
CDDrive(const std::string& device) :
        error(STATUS_NONE),
        cd_fd(-1)
{
    struct mntent *mnt;
    FILE *fp;

    error=STATUS_NO_SETMNTENT;
    if((fp = setmntent (MOUNTED, "r")) == NULL) return;

    error=STATUS_DEVICE_MOUNTED;
    while((mnt = getmntent (fp)) != NULL) {
        if(strcmp(mnt->mnt_fsname, device.c_str()) == 0) {
            endmntent(fp);
            return;
        }
    }
    endmntent(fp);
    
    error=STATUS_NO_DEVICE;
    if((cd_fd=open(device.c_str(), O_RDONLY | O_NONBLOCK))<0) return;

    error=STATUS_NONE;
    update_disc();
}

CDDrive::
~CDDrive()
{
    if(cd_fd>=0) {
        stop();
        close(cd_fd);
    }
}

void 
CDDrive::
play(const CDRelPos& pos)
{
    if(!disc.valid(pos)) return;

    Status state=status();
    if(state==STATUS_PAUSE) {
        resume();
        state=status();
    }
    if(state==STATUS_STOP || state==STATUS_PLAY) {
        CDAbsPos abspos_start=disc.abs_pos(pos);
        CDAbsPos abspos_end=disc.end();
        
        struct cdrom_msf msf;
        msf.cdmsf_min0=abspos_start.minute;
        msf.cdmsf_sec0=abspos_start.second;
        msf.cdmsf_frame0=abspos_start.frame;
        
        msf.cdmsf_min1=abspos_end.minute;
        msf.cdmsf_sec1=abspos_end.second;
        msf.cdmsf_frame1=abspos_end.frame;
        if(ioctl(cd_fd, CDROMSTART)) return;
        ioctl(cd_fd, CDROMPLAYMSF, &msf);
    }
}

Status
CDDrive::
status()
{
    if(error!=STATUS_NONE) return error;

    switch(ioctl(cd_fd, CDROM_DRIVE_STATUS, CDSL_CURRENT)) {
        case CDS_NO_INFO:
            disc.clear();
            return STATUS_ERROR;
        case CDS_NO_DISC:
        case CDS_TRAY_OPEN:
        case CDS_DRIVE_NOT_READY:
            disc.clear();
            return STATUS_NO_DISC;
        case CDS_DISC_OK:
            switch(ioctl(cd_fd, CDROM_DISC_STATUS)) {
                case CDS_NO_INFO:
                    disc.clear();
                    return STATUS_ERROR;
                case CDS_NO_DISC:
                    disc.clear();
                    return STATUS_NO_DISC;
                case CDS_DATA_1:
                case CDS_XA_2_1:
                case CDS_XA_2_2:
                    disc.clear();
                    return STATUS_DATA_DISC;
                case CDS_AUDIO:
                case CDS_MIXED:
                    if(ioctl(cd_fd, CDROM_MEDIA_CHANGED, CDSL_CURRENT))
                        update_disc();
                    struct cdrom_subchnl sc;
                    sc.cdsc_format=CDROM_MSF;
                    if(ioctl(cd_fd, CDROMSUBCHNL,&sc)) {
                        disc.clear();
                        return STATUS_ERROR;
                    }
                    switch(sc.cdsc_audiostatus) {
                        case CDROM_AUDIO_PLAY:
                            return STATUS_PLAY;
                        case CDROM_AUDIO_PAUSED:
                            return STATUS_PAUSE;
                        case CDROM_AUDIO_COMPLETED:
                        case CDROM_AUDIO_NO_STATUS:
                            return STATUS_STOP;
                        case CDROM_AUDIO_INVALID:
                        case CDROM_AUDIO_ERROR:
                            disc.clear();
                            return STATUS_ERROR;
                        default:
                            disc.clear();
                            return STATUS_ERROR;
                    }
                default:
                    disc.clear();
                    return STATUS_ERROR;
            }
        default:
            disc.clear();
            return STATUS_ERROR;
    }
}

unsigned int
CDDrive::
no_of_tracks()
{
    return disc.no_of_tracks();
}

unsigned int
CDDrive::
id()
{
    return disc.id();
}

CDRelPos
CDDrive::
pos()
{
    CDRelPos ret;
    ret.clear();

    struct cdrom_subchnl sc;
    sc.cdsc_format=CDROM_MSF;
    if(ioctl(cd_fd, CDROMSUBCHNL, &sc))
        return ret;

    if(sc.cdsc_audiostatus==CDROM_AUDIO_PLAY ||
       sc.cdsc_audiostatus==CDROM_AUDIO_PAUSED) {
        ret.set(sc.cdsc_trk,
                sc.cdsc_reladdr.msf.minute,
                sc.cdsc_reladdr.msf.second,
                sc.cdsc_reladdr.msf.frame);
    }
    return ret;
}

void
CDDrive::
pause()
{
    ioctl(cd_fd, CDROMPAUSE);
}

void
CDDrive::
resume()
{
    ioctl(cd_fd, CDROMRESUME);
}
 
void
CDDrive::
stop()
{
    ioctl(cd_fd, CDROMSTOP);
}

void
CDDrive::
eject()
{
    stop();
    ioctl(cd_fd, CDROMEJECT);
}

void
CDDrive::
close_tray()
{
    ioctl(cd_fd, CDROMCLOSETRAY);
}

CDRelPos
CDDrive::
add_pos(const CDRelPos& base, const CDAbsPos& offset)
{
    return disc.rel_pos(disc.abs_pos(base)+offset);
}

bool
CDDrive::
toc_changed()
{
    return disc.changed();
}
