//     This file is part of cdloop
//     Copyright (C) 2001-2003 Claus Brunzema <mail@cbrunzema.de>
// 
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
// 
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
// 
//     You should have received a copy of the GNU General Public License
//     along with this program; if not, write to the Free Software
//     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// 
#include "CDPlayer.hh"
#include <cstdio>
#include <unistd.h>
#include <wait.h>
#include <fstream>

static long
timeval_diff(const struct timeval& first, const struct timeval& last)
{
    return (last.tv_usec-first.tv_usec)+(last.tv_sec-first.tv_sec)*1000000;
}

Status
CDPlayer::
player_status()
{
    Status ret=drive.status();
    if(ret==STATUS_PLAY && loop.valid()) {
        ret=STATUS_LOOP;
    }
    if(ret==STATUS_PAUSE && gap_start.tv_sec>=0L && gap_start.tv_usec>=0L) {
        ret=STATUS_GAP;
    }
    return ret;
}

void
CDPlayer::
clear_gap_start()
{
    gap_start.tv_sec=-1L;
    gap_start.tv_usec=-1L;
}

void
CDPlayer::
clear_loop_vars()
{
    loop.clear();
    clear_gap_start();
}

void
CDPlayer::
finish_gap()
{
    if(player_status()==STATUS_GAP) 
        drive.play(loop.start);
    clear_gap_start();
}

void
CDPlayer::
controller()
{
    clear_loop_vars();
    shm->cmd=CMD_NO_CMD;
    while(shm->cmd!=CMD_QUIT) {
        if(player_status()==STATUS_LOOP) {
             if(drive.pos()>=loop.end) {
                if(loop.gap==0L) {
                    drive.play(loop.start);
                } else {
                    drive.pause();
                    gettimeofday(&gap_start, 0);
                }
            }
        } else if(player_status()==STATUS_GAP) {
            struct timeval tvbuffer;
            gettimeofday(&tvbuffer, 0);
            if(timeval_diff(gap_start, tvbuffer)>=loop.gap)
                finish_gap();
        }
                
        switch(shm->cmd) {
            case CMD_NO_CMD:
                usleep(6600);
                break;
            case CMD_PLAY:
                finish_gap();
                drive.play(shm->cmd_data.pos);
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_SET_LOOP:
                if(shm->cmd_data.loop.valid())
                    loop=shm->cmd_data.loop;
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_CLEAR_LOOP:
                finish_gap();
                clear_loop_vars();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_GET_ELAPSED_GAP:
                if(player_status()!=STATUS_GAP) {
                    shm->resp_data.elapsed_gap=0L;
                } else {
                    struct timeval tvbuffer;
                    gettimeofday(&tvbuffer, 0);
                    shm->resp_data.elapsed_gap=timeval_diff(gap_start,
                                                            tvbuffer);
                }
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_GET_STATUS:
                shm->resp_data.status=player_status();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_GET_NO_OF_TRACKS:
                shm->resp_data.no_of_tracks=drive.no_of_tracks();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_GET_ID:
                shm->resp_data.id=drive.id();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_GET_POS:
                shm->resp_data.pos=drive.pos();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_PAUSE:
                finish_gap();
                drive.pause();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_RESUME:
                drive.resume();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_STOP:
                drive.stop();
                clear_loop_vars();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_EJECT:
                drive.eject();
                clear_loop_vars();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_CLOSE_TRAY:
                drive.close_tray();
                clear_loop_vars();
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_ADD_POS:
                shm->resp_data.pos=drive.add_pos(shm->cmd_data.add_pos.base,
                                                 shm->cmd_data.add_pos.offset);
                shm->cmd=CMD_NO_CMD;
                break;
            case CMD_GET_TOC_CHANGED:
                shm->resp_data.toc_changed=drive.toc_changed();
                shm->cmd=CMD_NO_CMD;
                break;
            default:
//                shm->cmd=CMD_NO_CMD;
                break;
        }
    }
    _exit(0);
}

void
CDPlayer::
wait_for_controller()
{
    while(shm->cmd!=CMD_NO_CMD)
    { }
}

void
CDPlayer::
send_command(Cmd cmd)
{
    wait_for_controller();
    shm->cmd=cmd;
}

CDPlayer::
CDPlayer(const std::string device) :
        error(STATUS_NO_SHARED_MEM),
        control_child(-1),
        shmid(-1),
        shm(0),
        drive(device)
{
    shmid=shmget(IPC_PRIVATE, sizeof(Shmblock), IPC_CREAT | 0700);
    if(shmid==-1) return;
    shmctl(shmid, SHM_LOCK, NULL);

    shm=(Shmblock *)shmat(shmid, NULL, 0);
    shm->cmd=CMD_INIT;

    error=STATUS_NO_FORK;
    if(!(control_child=fork()))
    {
        controller();
    }
    if(control_child!=-1) error=STATUS_NONE;
}

CDPlayer::
~CDPlayer()
{
    if(control_child!=-1) {
        send_command(CMD_QUIT);
        int dummy;
        wait(&dummy);
    }
    if(shmid!=-1) {
        shmdt(shm);                                                          
        shmctl(shmid, IPC_RMID, NULL);
    }
}

void
CDPlayer::
play(const CDRelPos& pos)
{
    wait_for_controller();
    shm->cmd_data.pos=pos;
    send_command(CMD_PLAY);
}

void
CDPlayer::
set_loop(const Loop& loop)
{
    wait_for_controller();
    shm->cmd_data.loop=loop;
    send_command(CMD_SET_LOOP);
}

void
CDPlayer::
clear_loop()
{
    send_command(CMD_CLEAR_LOOP);
}

long
CDPlayer::
elapsed_gap()
{
    send_command(CMD_GET_ELAPSED_GAP);
    wait_for_controller();
    return shm->resp_data.elapsed_gap;
}

Status
CDPlayer::
status()
{
    if(error!=STATUS_NONE) return error;
    send_command(CMD_GET_STATUS);
    wait_for_controller();
    return shm->resp_data.status;
}

unsigned int
CDPlayer::
no_of_tracks()
{
    send_command(CMD_GET_NO_OF_TRACKS);
    wait_for_controller();
    return shm->resp_data.no_of_tracks;
}

unsigned int
CDPlayer::
id()
{
    send_command(CMD_GET_ID);
    wait_for_controller();
    return shm->resp_data.id;
}

CDRelPos
CDPlayer::
pos()
{
    send_command(CMD_GET_POS);
    wait_for_controller();
    return CDRelPos(shm->resp_data.pos);
}

void
CDPlayer::
pause()
{
    send_command(CMD_PAUSE);
}

void
CDPlayer::
resume()
{
    send_command(CMD_RESUME);
}

void
CDPlayer::
stop()
{
    send_command(CMD_STOP);
}

void
CDPlayer::
eject()
{
    send_command(CMD_EJECT);
}

void
CDPlayer::
close_tray()
{
    send_command(CMD_CLOSE_TRAY);
}

CDRelPos
CDPlayer::
add_pos(const CDRelPos& base, const CDAbsPos& offset)
{
    wait_for_controller();
    shm->cmd_data.add_pos.base=base;
    shm->cmd_data.add_pos.offset=offset;
    send_command(CMD_ADD_POS);
    wait_for_controller();
    return CDRelPos(shm->resp_data.pos);
}
    
bool
CDPlayer::
toc_changed()
{
    send_command(CMD_GET_TOC_CHANGED);
    wait_for_controller();
    return shm->resp_data.toc_changed;
}
    
